import os
import time
import awkward as ak
import numpy as np
import pandas as pd
import warnings

import epix


def main(args, return_df=False, return_wfsim_instructions=False, strax=False):
    """Call this function from the run_epix script"""

    if args['debug']:
        print("epix configuration: ", args)
        # TODO: also add memory information (see straxer) and change this to debug
        # Getting time information:
        starttime = time.time()
        tnow = starttime

    # Loading data:
    inter, n_simulated_events = epix.loader(args['path'],
                                            args['file_name'],
                                            args['debug'],
                                            outer_cylinder=args['outer_cylinder'],
                                            kwargs_uproot_arrays={'entry_start': args['entry_start'],
                                                                  'entry_stop': args['entry_stop']},
                                            cut_by_eventid=args['cut_by_eventid']
                                            )

    if args['debug']:
        tnow = monitor_time(tnow, 'load data.')
        print(f"Finding clusters of interactions with a dr = {args['micro_separation']} mm"
               f" and dt = {args['micro_separation_time']} ns")

    # Cluster finding and clustering (convert micro_separation mm -> cm):
    inter = epix.find_cluster(inter, args['micro_separation']/10, args['micro_separation_time'])

    if args['debug']:
        tnow = monitor_time(tnow, 'find clusters.')

    result = epix.cluster(inter, args['tag_cluster_by'] == 'energy')

    if args['debug']:
        tnow = monitor_time(tnow, 'merge clusters.')

    # Add eventid again:
    result['evtid'] = ak.broadcast_arrays(inter['evtid'][:, 0], result['ed'])[0]

    # Sort detector volumes and keep interactions in selected ones:
    if args['debug']:
        print('Removing clusters not in volumes:', *[v.name for v in args['detector_config']])
        print(f'Number of clusters before: {np.sum(ak.num(result["ed"]))}')

    # Returns all interactions which are inside in one of the volumes,
    # Checks for volume overlap, assigns Xe density and create_S2 to
    # interactions. EField comes later since interpolated maps cannot be
    # called inside numba functions.
    res_det = epix.in_sensitive_volume(result, args['detector_config'])

    # Adding new fields to result:
    for field in res_det.fields:
        result[field] = res_det[field]
    m = result['vol_id'] > 0  # All volumes have an id larger zero
    result = result[m]

    # Removing now empty events as a result of the selection above:
    m = ak.num(result['ed']) > 0
    result = result[m]

    if args['debug']:
        print(f'Number of clusters after: {np.sum(ak.num(result["ed"]))}')
        print('Assigning electric field to clusters')

    if not ak.any(m):
        # There are not any events left so return empty array:
        #TODO: Add option for WFSim
        warnings.warn('No interactions left, return empty DataFrame.')
        if return_df:
            if args.output_path and not os.path.isdir(args.output_path):
                os.makedirs(args.output_path)

            output_path_and_name = os.path.join(args.output_path,
                                                args['file_name'][:-5] + "_wfsim_instructions.csv")
            df = pd.DataFrame()
            df.to_csv(output_path_and_name, index=False)
            return

    # Add electric field to array:
    efields = np.zeros(np.sum(ak.num(result)), np.float32)
    # Loop over volume and assign values:
    for volume in args['detector_config']:
        if isinstance(volume.electric_field, (float, int)):
            ids = epix.awkward_to_flat_numpy(result['vol_id'])
            m = ids == volume.volume_id
            efields[m] = volume.electric_field
        else:
            efields = volume.electric_field(epix.awkward_to_flat_numpy(result.x),
                                            epix.awkward_to_flat_numpy(result.y),
                                            epix.awkward_to_flat_numpy(result.z)
                                            )

    result['e_field'] = epix.reshape_awkward(efields, ak.num(result))

    # Sort entries (in an event) by in time, then chop all delayed
    # events which are too far away from the rest.
    # (This is a requirement of WFSim)
    result = result[ak.argsort(result['t'])]
    dt = result['t'] - result['t'][:, 0]
    result = result[dt <= args['max_delay']]

    if args['debug']:
        print('Generating photons and electrons for events')
    # Generate quanta:
    photons, electrons = epix.quanta_from_NEST(epix.awkward_to_flat_numpy(result['ed']),
                                               epix.awkward_to_flat_numpy(result['nestid']),
                                               epix.awkward_to_flat_numpy(result['e_field']),
                                               epix.awkward_to_flat_numpy(result['A']),
                                               epix.awkward_to_flat_numpy(result['Z']),
                                               epix.awkward_to_flat_numpy(result['create_S2']),
                                               density=epix.awkward_to_flat_numpy(result['xe_density']))
    result['photons'] = epix.reshape_awkward(photons, ak.num(result['ed']))
    result['electrons'] = epix.reshape_awkward(electrons, ak.num(result['ed']))
    if args['debug']:
        _ = monitor_time(tnow, 'get quanta.')

    # Separate events in time
    number_of_events = len(result["t"])
    if args['source_rate'] == -1:
        # Only needed for a clean separation:
        result['t'] = result['t'] - result['t'][:, 0]

        dt = epix.times_for_clean_separation(number_of_events, args['max_delay'])
        if args['debug']:
            print('Clean event separation')
    elif args['source_rate'] == 0:
        # In case no delay should be applied we just add zeros
        dt = np.zeros(number_of_events)
    else:
        # Rate offset computed based on the specified rate and job_id.
        # Assumes all jobs were started with the same number of events.
        offset = (args['job_number']*n_simulated_events)/args['source_rate']
        dt = epix.times_from_fixed_rate(args['source_rate'],
                                        number_of_events,
                                        n_simulated_events,
                                        offset
                                        )
        if args['debug']:
            print(f"Fixed event rate of {args['source_rate']} Hz")

    result['t'] = result['t'][:, :] + dt

    # Reshape instructions:
    instructions = epix.awkward_to_wfsim_row_style(result)
    if args['source_rate'] != 0:
        # Only sort by time again if source rates were applied, otherwise
        # things are already sorted within the events and should stay this way.
        instructions = np.sort(instructions, order='time')

    ins_df = pd.DataFrame(instructions)

    if return_df:
        if args['output_path'] and not os.path.isdir(args['output_path']):
            os.makedirs(args['output_path'])

        output_path_and_name = os.path.join(args['output_path'], args['file_name'][:-5] + "_wfsim_instructions.csv")
        if os.path.isfile(output_path_and_name):
            warnings.warn("Output file already exists - Overwriting")
        ins_df.to_csv(output_path_and_name, index=False)

        print('Done')
        print('Instructions saved to ', output_path_and_name)
        if args['debug']:
            _ = monitor_time(starttime, 'run epix.')

    if return_wfsim_instructions:
        return instructions


def monitor_time(prev_time, task):
    t = time.time()
    print(f'It took {(t - prev_time):.4f} sec to {task}')
    return t


def setup(args):
    """
    Function which sets-up configurations. (for strax conversion)
    Is returning the dict like this nessecairy?
    :return:
    """
    # Getting file path and split it into directory and file name:
    if not ('path' in args and 'file_name' in args):
        p = args['input_file']
        if '/' in p:
            p = p.split('/')
        else:
            p = ["", p]
        if p[0] == "":
            p[0] = "/"
        args['path'] = os.path.join(*p[:-1])
        args['file_name'] = p[-1]

    # Init detector volume according to settings and get outer_cylinder
    # for data reduction of non-relevant interactions.
    args['detector_config'] = epix.init_detector(args['detector'].lower(), args['detector_config_override'])
    outer_cylinder = getattr(epix.detectors, args['detector'].lower())
    _, args['outer_cylinder'] = outer_cylinder()
    return args



import numpy as np
import pandas as pd
import numba
import awkward as ak
from .common import reshape_awkward
from sklearn.cluster import DBSCAN

def find_cluster(interactions, cluster_size_space, cluster_size_time):
    """
    Function which finds cluster within a event.

    Args:
        x (pandas.DataFrame): Subentries of event must contain the
            fields, x,y,z,time
        cluster_size_space (float): Max spatial distance between two points to
            be inside a cluster [cm].
        cluster_size_time (float): Max time distance between two points to be 
            inside a cluster [ns].
    
    Returns:
        awkward.array: Adds to interaction a cluster_ids record.
    """
    # TODO is there a better way to get the df?
    df = []
    for key in ['x', 'y', 'z', 'ed', 't']:
        df.append(ak.to_pandas(interactions[key], anonymous=key))
    df = pd.concat(df, axis=1)

    # Splitting into individual events and apply time clustering:
    groups = df.groupby('entry')
    df["time_cluster"] = np.concatenate(groups.apply(lambda x: simple_1d_clustering(x.t.values, cluster_size_time)))

    # Splitting into individual events and time cluster and apply space clustering space:
    groups = df.groupby(['entry', 'time_cluster'])
    groups = groups.apply(lambda x: _find_cluster(x, cluster_size_space))

    for i in np.unique(groups.index.get_level_values(0)):
        add_to_cluster = 0
        for j in range(len(groups[i])):
            groups[i][j]+=add_to_cluster
            add_to_cluster = np.max(groups[i][j])+1

    df['cluster_id'] = np.concatenate(groups.values)

    ci = df.loc[:, 'cluster_id'].values
    offsets = ak.num(interactions['x'])
    interactions['cluster_ids'] = reshape_awkward(ci, offsets)
    
    return interactions

@numba.jit(nopython=True)
def simple_1d_clustering(data, scale):
    """
    Function to cluster one dimensional data.

    Args:
        data (numpy.array): one dimensional array to be clusterd
        scale (float): Max distance between two points to
            be inside a cluster.
    
    Returns:
        clusters_undo_sort (np.array): Cluster Labels
    """
    
    idx_sort = np.argsort(data)
    idx_undo_sort = np.argsort(idx_sort)

    data_sorted = data[idx_sort]

    diff = data_sorted[1:] - data_sorted[:-1]
    
    clusters = [0]
    c = 0
    for value in diff:
        if value <= scale:
            clusters.append(c)
        elif value > scale:
            c=c+1
            clusters.append(c)

    clusters_undo_sort = np.array(clusters)[idx_undo_sort]
    
    return clusters_undo_sort

def _find_cluster(x, cluster_size_space):
    """
    Function which finds cluster within a event.
    Args:
        x (pandas.DataFrame): Subentries of event must contain the
            fields, x,y,z,time
        cluster_size_space (float): Max distance between two points to
            be inside a cluster [cm].
    Returns:
        functon: to be used in groupby.apply.
    """
    db_cluster = DBSCAN(eps=cluster_size_space, min_samples=1)
    xprime = x[['x', 'y', 'z']].values 
    return db_cluster.fit_predict(xprime)


def cluster(inter, classify_by_energy=False):
    """
    Function which clusters the found clusters together.

    To cluster events a weighted mean is computed for time and position.
    The individual interactions are weighted by their energy.
    The energy of clustered interaction is given by the total sum.
    Events can be classified either by the first interaction in time in the
    cluster or by the highest energy deposition.

    Args:
        inter (awkward.Array): Array containing at least the following
            fields: x,y,z,t,ed,cluster_ids, type, parenttype, creaproc,
            edproc.

    Kwargs:
        classify_by_energy (bool): If true events are classified
            according to the properties of the highest energy deposit
            within the cluster. If false cluster is classified according
            to first interaction.

    Returns:
        awkward.Array: Clustered events with nest conform
            classification.
    """
    # Sort interactions by cluster_ids to simplify looping
    inds = ak.argsort(inter['cluster_ids'])
    inter = inter[inds]

    # TODO: Better way to do this with awkward?
    x = inter['x']
    y = inter['y']
    z = inter['z']
    ed = inter['ed']
    time = inter['t']
    ci = inter['cluster_ids']
    types = inter['type']
    parenttype = inter['parenttype']
    creaproc = inter['creaproc']
    edproc = inter['edproc']

    # Init result and cluster:
    res = ak.ArrayBuilder()
    _cluster(x, y, z, ed, time, ci,
             types, parenttype, creaproc, edproc,
             classify_by_energy, res)
    return res.snapshot()


@numba.njit
def _cluster(x, y, z, ed, time, ci,
             types, parenttype, creaproc, edproc,
             classify_by_energy, res):
    # Loop over each event
    nevents = len(ed)
    for ei in range(nevents):
        # Init a new list for clustered interactions within event:
        res.begin_list()

        # Init buffers:
        ninteractions = len(ed[ei])
        x_mean = 0
        y_mean = 0
        z_mean = 0
        t_mean = 0
        ed_tot = 0

        current_ci = 0  # Current cluster id
        i_class = 0  # Index for classification (depends on users requirement)
        # Set classifier start value according to user request, interactions
        # are classified either by
        if classify_by_energy:
            # Highest energy
            classifier_max = 0
        else:
            # First interaction
            classifier_max = np.inf

        # Loop over all interactions within event:
        for ii in range(ninteractions):
            if current_ci != ci[ei][ii]:
                # Cluster Id has changed compared to previous interaction,
                # hence we have to write out our result and empty the buffer,
                # but first classify event:
                A, Z, nestid = classify(types[ei][i_class],
                                        parenttype[ei][i_class],
                                        creaproc[ei][i_class],
                                        edproc[ei][i_class])

                # Write result, simple but extensive with awkward...
                _write_result(res, x_mean, y_mean, z_mean,
                              ed_tot, t_mean, A, Z, nestid)

                # Update cluster id and empty buffer
                current_ci = ci[ei][ii]
                x_mean = 0
                y_mean = 0
                z_mean = 0
                t_mean = 0
                ed_tot = 0

                # Reset classifier:
                if classify_by_energy:
                    classifier_max = 0
                else:
                    classifier_max = np.inf

            # We have to gather information of current cluster:
            e = ed[ei][ii]
            t = time[ei][ii]
            x_mean += x[ei][ii] * e
            y_mean += y[ei][ii] * e
            z_mean += z[ei][ii] * e
            t_mean += t * e
            ed_tot += e

            if classify_by_energy:
                # In case we want to classify the event by energy.
                if e > classifier_max:
                    i_class = ii
                    classifier_max = e
            else:
                # or by first arrival time:
                if t < classifier_max:
                    i_class = ii
                    classifier_max = t

        # Before we are done with this event we have to classify and
        # write the last interaction
        A, Z, nestid = classify(types[ei][i_class],
                                parenttype[ei][i_class],
                                creaproc[ei][i_class],
                                edproc[ei][i_class])
        _write_result(res, x_mean, y_mean, z_mean,
                      ed_tot, t_mean, A, Z, nestid)

        res.end_list()


classifier = np.zeros(7, dtype=[(('Interaction type', 'types'), np.dtype('<U30')),
                                (('Interaction type of the parent', 'parenttype'), np.dtype('<U30')),
                                (('Creation process', 'creaproc'), np.dtype('<U30')),
                                (('Energy deposit process', 'edproc'), np.dtype('<U30')),
                                (('Atomic mass number', 'A'), np.int16),
                                (('Atomic number', 'Z'), np.int16),
                                (('Nest Id for qunata generation', 'nestid'), np.int16)]
                      )
classifier['types'] = ['None', 'neutron', 'alpha', 'None','None', 'gamma', 'e-']
classifier['parenttype'] = ['None', 'None', 'None', 'Kr83[9.405]','Kr83[41.557]', 'None', 'None']
classifier['creaproc'] = ['None', 'None', 'None', 'None', 'None','None', 'None']
classifier['edproc'] = ['ionIoni', 'hadElastic', 'None', 'None','None', 'None', 'None']
classifier['A'] = [0, 0, 4, 0, 0, 0, 0]
classifier['Z'] = [0, 0, 2, 0, 0, 0, 0]
classifier['nestid'] = [0, 0, 6, 11, 11, 7, 8]


@numba.njit
def classify(types, parenttype, creaproc, edproc):
    for c in classifier:
        m = 0
        m += (c['types'] == types) or (c['types'] == 'None')
        m += (c['parenttype'] == parenttype) or (c['parenttype'] == 'None')
        m += (c['creaproc'] == creaproc) or (c['creaproc'] == 'None')
        m += (c['edproc'] == edproc) or (c['edproc'] == 'None')

        if m == 4:
            return c['A'], c['Z'], c['nestid']

    # If our data does not match any classification make it a nest None type
    # TODO: fix me
    return 0, 0, 12


@numba.njit
def _write_result(res, x_mean, y_mean, z_mean,
                  ed_tot, t_mean, A, Z, nestid):
    """
    Helper to write result into record array.
    """
    res.begin_record()
    res.field('x')
    res.real(x_mean / ed_tot)
    res.field('y')
    res.real(y_mean / ed_tot)
    res.field('z')
    res.real(z_mean / ed_tot)
    res.field('t')
    res.real(t_mean / ed_tot)
    res.field('ed')
    res.real(ed_tot)
    res.field('nestid')
    res.integer(nestid)
    res.field('A')
    res.integer(A)
    res.field('Z')
    res.integer(Z)
    res.end_record()



import numba
import numpy as np
import awkward as ak


def reshape_awkward(array, offset):
    """
    Function which reshapes array according to a list of offsets. Only
    works for a single jagged layer.

    Args:
        array: Flatt array which should be jagged.
        offset: Length of subintervals


    Returns:
        res: awkward1.ArrayBuilder object.
    """
    res = ak.ArrayBuilder()
    _reshape_awkward(array, offset, res)
    return res.snapshot()


@numba.njit
def _reshape_awkward(array, offsets, res):
    start = 0
    end = 0
    for o in offsets:
        end += o
        res.begin_list()
        for value in array[start:end]:
            res.real(value)
        res.end_list()
        start = end


def awkward_to_flat_numpy(array):
    return (ak.to_numpy(ak.flatten(array)))


@numba.njit
def mulit_range(offsets):
    res = np.zeros(np.sum(offsets), dtype=np.int32)
    i = 0
    for o in offsets:
        res[i:i+o] = np.arange(0, o, dtype=np.int32)
        i += o
    return res

@numba.njit
def offset_range(offsets):
    """
    Computes range of constant event ids while in same offset. E.g.
    for an array [1], [1,2,3], [5] this function yields [0, 1, 1, 1, 2].

    Args:
        offsets (ak.array): jagged array offsets.

    Returns:
        np.array: Indicies.
    """
    res = np.zeros(np.sum(offsets), dtype=np.int32)
    i = 0
    for ind, o in enumerate(offsets):
        res[i:i+o] = ind
        i += o
    return res


__version__ = '0.1.0'

from .common import *
from .io import *
from .clustering import *
from .quanta_generation import *
from .detector_volumes import *
from .clustering import *
from .electric_field_handler import *
from .event_separation import *
from .detectors import *
from .run_epix import *



import inspect
import numba
import numpy as np
import awkward as ak
import epix


def init_detector(detector_name, config_file):
    """
    Function which initializes the detector. Reads in config and
    overwrites default settings.

    :param detector_name: Detector as defined in detector.py.
    :param config_file: File to be used to customize the default
        settings e.g. electric field.

    :returns: A list of volumes which represent the detector.
    """
    if not hasattr(epix.detectors, detector_name):
        raise ValueError('The specified "detector_name" must be defined in "epix.detectors". '
                         f'Was not able to find {detector_name}.')

    # Getting default volume object by name:
    volumes = getattr(epix.detectors, detector_name)
    volumes, _ = volumes()

    if config_file:
        # Load config and overwrite default settings for volumes.
        config = epix.io.load_config(config_file)

        # Update default setting with new settings:
        for volume_name, options in config.items():
            if volume_name in volumes:
                default_options = volumes[volume_name]
            else:
                raise ValueError(f'Cannot find "{volume_name}" among the volumes to be initialized.'
                                 f' Valid volumes are: {[k for k in volumes.keys()]}')
            for key, values in options.items():
                default_options[key] = values

    # Now loop over default options and init volumes:
    detector = []
    args_volumes = inspect.signature(epix.SensitiveVolume).parameters.keys()
    for volume_name, default_options in volumes.items():
        if not default_options['to_be_stored']:
            # This volume should be skipped
            continue

        kwargs = {}
        for option_key, option_value in default_options.items():
            if option_key not in args_volumes:
                # This option is not needed to init the volume
                continue

            if option_key == 'electric_field' and isinstance(option_value, str):
                # Special case we have to init the efield first:
                efield = epix.MyElectricFieldHandler(option_value)
                option_value = lambda x, y, z: efield.get_field(x, y, z,
                                                                outside_map=default_options['efield_outside_map']
                                                                )
            kwargs[option_key] = option_value

        detector.append(epix.SensitiveVolume(name=volume_name, **kwargs))

    # Test if volume name and id are unique:
    names = [volume.name for volume in detector]
    ids = [volume.volume_id for volume in detector]
    if len(np.unique(names)) != len(names) or len(np.unique(ids)) != len(ids):
        raise ValueError(f'Detector volumes must have unique names and ids! Found {names} and {ids}.')

    return detector


class SensitiveVolume:
    def __init__(self, name, volume_id, roi, electric_field, create_S2, xe_density=2.862):
        """
        Sensitive detector volume for which S1 and/or S2 signals should be
        generated.
        
        Args:
            name (str): Name of the volume.
            vol_id (int): Id of the volume, must be unique and greater
                zero. Zero is reserved for event which are not in any
                volume.
            roi (function): Function which takes x, y, z as arguments.
                Must return True or False. Must be numba njit-able.
            e_field (function): Interpolated efield map associated
                with volume.
            create_S2 (bool): Indicates if electrons should be generated
                for wfsim.

        Kwargs:
            xe_properties (dict): Dictionary of additional xenon
                properties which should be forwarded to nest.
        """
        self.name = name
        self.volume_id = volume_id
        self.roi = roi

        self.electric_field = electric_field
        self.xe_density = xe_density
        self.create_S2 = create_S2
        self._is_valid()

    def _is_valid(self):
        """
        Function which tests if the inputs are valid.
        """
        # Test vol_id:
        assert isinstance(self.volume_id, int), ('The volume id vol_id must be an '
                                                 f'integer, but {self.volume_id} was '
                                                 'given.')
        assert self.volume_id > 0, ('The volume id vol_id must be greater zero, '
                                    f'but {self.volume_id} was given.')

        # Test if ROI function is defined properly:
        assert callable(self.roi), ('roi must be a callable function '
                                    'which depends on x,y,z.')

        # Testing the electric field:
        if not (callable(self.electric_field) or
                isinstance(self.electric_field, (int, float))):
            raise ValueError('e_field must be either a function or '
                             'a constant!')

        if callable(self.electric_field):
            args = inspect.getfullargspec(self.electric_field).args
            m = np.all(np.isin(['x', 'y', 'z'], args))
            m = m & (len(args) == 3)
            assert m, ('Wrong arguments for e_field. Expected arguments: '
                       f'"x", "y" and "z" but {args} were given.')
        # Cannot add a specific if **kwargs are valid properties. Cannot
        # inspect nestpy functions.


@numba.njit
def in_cylinder(x, y, z, min_z, max_z, max_r):
    """
    Function which checks if a given set of coordinates is within the
    boundaries of the specified cylinder. 
    
    Args:
        x,y,z: Coordinates of the interaction
        min_z: Inclusive lower z boundary
        max_z: Exclusive upper z boundary
        max_r: Exclusive radial boundary
    """
    r = np.sqrt(x**2 + y**2)
    m = r < max_r
    m = m & (z < max_z)
    m = m & (z >= min_z)
    return m


def in_sensitive_volume(events, sensitive_volumes):
    """
    Function which identifies which events are inside sensitive volumes.

    Further, tests if sensitive volumes overlap. Assigns volume id, and
    xenon density to interactions.

    Args:
        events (ak.records): Awkward record of the interactions.
        sensitive_volumes (dict): Dictionary of the different volumes
            defined via the SensitiveVolume class.

    Returns:
        ak.array: Awkward array containing the event ids.
    """
    for ind, vol in enumerate(sensitive_volumes):
        res = ak.ArrayBuilder()
        res = _inside_sens_vol(events['x'],
                               events['y'],
                               events['z'],
                               vol.roi,
                               vol.volume_id,
                               vol.xe_density,
                               vol.create_S2,
                               res)
        if ind:
            # Now we add the other results, but first test if 
            # volumes overlap. Only possible if we explicitly loop
            # over everything. This reduces performance but adds layer of
            # safety.
            m = (result['vol_id'] > 0) & (res['vol_id'] == vol.volume_id)
            if ak.any(m):
                overlapping_id = result[m][0]
                # Get volume name:
                name = [vol.name for vol in sensitive_volumes if vol.volume_id == overlapping_id][0]
                raise ValueError(f'The volume {vol.name} is overlapping with'
                                 f' volume {name}!')
            new_results = res.snapshot()
            for field in result.fields:
                # Workaround since we cannot sum up records-arrays anymore
                result[field] = result[field] + new_results[field]
        else:
            # First result initiates the array
            result = res.snapshot()
    return result


@numba.njit()
def _inside_sens_vol(xp, yp, zp, roi, vol_id, vol_density, create_S2, res):
    nevents = len(xp)
    for i in range(nevents):
        # Loop over all events
        res.begin_list()
        nint = len(xp[i])
        if nint:
            for j in range(nint):
                # Loop over all interactions within an event.
                res.begin_record()
                if roi(x=xp[i][j], y=yp[i][j], z=zp[i][j]):
                    res.field('xe_density')
                    res.real(vol_density)
                    res.field('vol_id')
                    res.integer(vol_id)
                    res.field('create_S2')
                    res.boolean(create_S2)
                else:
                    res.field('vol_id')
                    res.integer(0)
                    res.field('xe_density')
                    res.real(0)
                    res.field('create_S2')
                    res.boolean(False)
                res.end_record()
        res.end_list()
    return res



from scipy.interpolate import RegularGridInterpolator as RGI
import pandas as pd
import numpy as np
import os
import gzip
import json


class MyElectricFieldHandler:
    def __init__(self, field_map=""):
        """
        The field map, defined over a regular grid, should be
        a .csv or .json.gz file. 

        Structure of the csv file:
        Columns "r" "z" and "E", with lenght in cm and field in V/cm.
        The elements are delimited by a ",". 

        Structure of the json.gz file:
        Contains the "r" and "z" coordinates in cm under the key
        "coordinate_system" and the field in V/cm under the key "map".
        """
        self.map = field_map
        if os.path.isfile(self.map):
            self._load_field()
            self._get_coordinates()
            self._build_interpolator()
        else:
            raise ValueError(f'Cannot open "{self.map}". It is not a valid file'
                             ' for the electirc field map.')

    def _load_field(self):
        file_ending = self.map.split(".")[-1]

        if file_ending == "csv":
            self.field = pd.read_csv(self.map)
        elif file_ending == "gz":
            with gzip.open(self.map, 'rb') as f:
                field_map = json.load(f)
            self.field = pd.DataFrame()
            self.field["r"] = np.array(field_map["coordinate_system"])[:,0]
            self.field["z"] = np.array(field_map["coordinate_system"])[:,1]
            self.field["E"] = np.array(field_map["map"])
        else:
            raise ValueError(f'Cannot open "{self.map}". File extension is not valid'
                             ' for the electric field map. Use .csv or .json.gz')

    def _get_coordinates(self):
        self.R = np.unique(self.field['r'])
        self.Z = np.unique(self.field['z'])

    def _build_interpolator(self):
        e_tmp = np.reshape(np.array(self.field.E),
                           (len(self.Z), len(self.R)))
        self.interpolator = RGI([self.Z, self.R],
                                e_tmp,
                                bounds_error=False,
                                fill_value=None)
    
    def get_field(self, x, y, z, outside_map=np.nan):
        """
        Function which returns the electric field at a certain position
        according to an efield map.

        Args:
            x (np.array): x coordinate of the interaction in cm
            y (np.array): y coordinate of the interaction in cm
            z (np.array): z coordinate of the interaction in cm

        Kwargs:
            outside_map (float): Default value to be used if interaction
                was not within the range of the map. Default np.nan
        :return:
        """
        r = np.sqrt(x**2+y**2)
        efield = self.interpolator((z, r))
        efield[np.isnan(efield)] = outside_map
        return efield



import numpy as np
import uproot
import awkward as ak

import os
import warnings
import wfsim
import configparser

from .common import awkward_to_flat_numpy, offset_range

SUPPORTED_OPTION = {'to_be_stored': 'getboolean',
                    'electric_field': ('getfloat', 'get'),
                    'create_S2': 'getboolean',
                    'xe_density': 'getfloat',
                    'efield_outside_map': 'getfloat',
                    }


def load_config(config_file_path):
    """
    Loads config file and returns dictionary.

    :param config_file_path:
    :return: dict
    """
    config = configparser.ConfigParser()
    config.read(config_file_path)
    sections = config.sections()
    if not len(sections):
        raise ValueError(f'Cannot load sections from config file "{config_file_path}".' 
                         'Have you specified a wrong file?')
    settings = {}
    for s in sections:
        options = {}
        c = config[s]
        for key in c.keys():
            if key not in SUPPORTED_OPTION:
                warnings.warn(f'Option "{key}" of section {s} is not supported'
                              ' and will be ignored.')
                continue
            # Get correct get method to convert string input:
            if key == 'electric_field':
                # Electric field is a bit more complicated can be
                # either a float or string:
                try:
                    getter = getattr(c, SUPPORTED_OPTION[key][0])
                    options[key] = getter(key)
                except ValueError:
                    getter = getattr(c, SUPPORTED_OPTION[key][1])
                    options[key] = getter(key)
            else:
                try:
                    getter = getattr(c, SUPPORTED_OPTION[key])
                    options[key] = getter(key)
                except Exception as e:
                    raise ValueError(f'Cannot load "{key}" from section "{s}" in config file.') from e

        settings[s] = options
    return settings


def loader(directory,
           file_name,
           arg_debug=False,
           outer_cylinder=None,
           kwargs_uproot_arrays={},
           cut_by_eventid=False,
           ):
    """
    Function which loads geant4 interactions from a root file via
    uproot4.

    Besides loading, a simple data selection is performed. Units are
    already converted into strax conform values. mm -> cm and s -> ns.

    Args:
        directory (str): Directory in which the data is stored.
        file_name (str): File name
        arg_debug (bool): If true, print out loading information.
        outer_cylinder (dict): If specified will cut all events outside of the
            given cylinder.
        kwargs_uproot_arrays (dict): Keyword arguments passed to .arrays of
            uproot4.
        cut_by_eventid (bool): If true event start/stop are applied to
            eventids, instead of rows.

    Returns:
        awkward1.records: Interactions (eventids, parameters, types).
        integer: Number of events simulated.
    """
    ttree, n_simulated_events = _get_ttree(directory, file_name)

    if arg_debug:
        print(f'Total entries in input file = {ttree.num_entries}')
        cutby_string='output file entry'
        if cut_by_eventid:
            cutby_string='g4 eventid'

        if kwargs_uproot_arrays['entry_start'] is not None:
            print(f'Starting to read from {cutby_string} {kwargs_uproot_arrays["entry_start"]}')
        if kwargs_uproot_arrays['entry_stop'] is not None:
            print(f'Ending read in at {cutby_string} {kwargs_uproot_arrays["entry_stop"]}')

    # If user specified entry start/stop we have to update number of
    # events for source rate computation:
    if kwargs_uproot_arrays['entry_start'] is not None:
        start = kwargs_uproot_arrays['entry_start']
    else:
        start = 0

    if kwargs_uproot_arrays['entry_stop'] is not None:
        stop = kwargs_uproot_arrays['entry_stop']
    else:
        stop = n_simulated_events
    n_simulated_events = stop - start

    if cut_by_eventid:
        # Start/stop refers to eventid so drop start drop from kwargs
        # dict if specified, otherwise we cut again on rows.
        kwargs_uproot_arrays.pop('entry_start', None)
        kwargs_uproot_arrays.pop('entry_stop', None)

    # Columns to be read from the root_file:
    column_names = ["x", "y", "z", "t", "ed",
                    "type", "trackid",
                    "parenttype", "parentid",
                    "creaproc", "edproc"]

    # Conversions and parameters to be computed:
    alias = {'x': 'xp/10',  # converting "geant4" mm to "straxen" cm
             'y': 'yp/10',
             'z': 'zp/10',
             'r': 'sqrt(x**2 + y**2)',
             't': 'time*10**9'
             }

    if outer_cylinder:
        cut_string = (f'(r < {outer_cylinder["max_r"]})'
                      f' & ((zp >= {outer_cylinder["min_z"] * 10}) & (zp < {outer_cylinder["max_z"] * 10}))')
    else:
        cut_string = None

    # Read in data, convert mm to cm and perform a first cut if specified:
    interactions = ttree.arrays(column_names,
                                cut_string,
                                aliases=alias,
                                **kwargs_uproot_arrays)
    eventids = ttree.arrays('eventid', **kwargs_uproot_arrays)
    eventids = ak.broadcast_arrays(eventids['eventid'], interactions['x'])[0]
    interactions['evtid'] = eventids

    if np.any(interactions['ed'] < 0):
        warnings.warn('At least one of the energy deposits is negative!')

    # Removing all events with zero energy deposit
    m = interactions['ed'] > 0
    if cut_by_eventid:
        # ufunc does not work here...
        m2 = (interactions['evtid'] >= start) & (interactions['evtid'] < stop)
        m = m & m2
    interactions = interactions[m]

    # Removing all events with no interactions:
    m = ak.num(interactions['ed']) > 0
    interactions = interactions[m]

    return interactions, n_simulated_events


def _get_ttree(directory, file_name):
    """
    Function which searches for the correct ttree in MC root file.

    :param directory: Directory where file is
    :param file_name: Name of the file
    :return: root ttree and number of simulated events
    """
    root_dir = uproot.open(os.path.join(directory, file_name))

    # Searching for TTree according to old/new MC file structure:
    if root_dir.classname_of('events') == 'TTree':
        ttree = root_dir['events']
        n_simulated_events = root_dir['nEVENTS'].members['fVal']
    elif root_dir.classname_of('events/events') == 'TTree':
        ttree = root_dir['events/events']
        n_simulated_events = root_dir['events/nbevents'].members['fVal']
    else:
        ttrees = []
        for k, v in root_dir.classnames().items():
            if v == 'TTree':
                ttrees.append(k)
        raise ValueError(f'Cannot find ttree object of "{file_name}".'
                         'I tried to search in events and events/events.'
                         f'Found a ttree in {ttrees}?')
    return ttree, n_simulated_events


# ----------------------
# Outputing wfsim instructions:
# ----------------------
int_dtype = wfsim.instruction_dtype


def awkward_to_wfsim_row_style(interactions):
    """
    Converts awkward array instructions into instructions required by
    WFSim.

    :param interactions: awkward.Array containing GEANT4 simulation
        information.
    :return: Structured numpy.array. Each row represents either a S1 or
        S2
    """
    ninteractions = np.sum(ak.num(interactions['ed']))
    res = np.zeros(2 * ninteractions, dtype=int_dtype)

    # TODO: Currently not supported rows with only electrons or photons due to
    # this super odd shape
    for i in range(2):
        res['event_number'][i::2] = offset_range(ak.to_numpy(ak.num(interactions['evtid'])))
        res['type'][i::2] = i + 1
        res['x'][i::2] = awkward_to_flat_numpy(interactions['x'])
        res['y'][i::2] = awkward_to_flat_numpy(interactions['y'])
        res['z'][i::2] = awkward_to_flat_numpy(interactions['z'])
        res['time'][i::2] = awkward_to_flat_numpy(interactions['t'])
        res['g4id'][i::2] = awkward_to_flat_numpy(interactions['evtid'])
        res['vol_id'][i::2] = awkward_to_flat_numpy(interactions['vol_id'])
        res['e_dep'][i::2] = awkward_to_flat_numpy(interactions['ed'])
        
        recoil = awkward_to_flat_numpy(interactions['nestid'])
        res['recoil'][i::2] = np.where(np.isin(recoil, [0,6,7,8,11]), recoil, 8)

        if i:
            res['amp'][i::2] = awkward_to_flat_numpy(interactions['electrons'])
        else:
            res['amp'][i::2] = awkward_to_flat_numpy(interactions['photons'])

    # Remove entries with no quanta
    res = res[res['amp'] > 0]
    return res



import numpy as np

def times_for_clean_separation(n_events, MaxDelay):
    """
    Function to generate evenly spaced event times.

    Args:
        n_events (int): Number of events 
        MaxDelay (float): Time difference between events. Should be large enough to 
            prevent event pile-up.

    Returns:
        dt (numpy.array): Array containing the start times of the events.
    """
    
    dt = np.arange(1, n_events+1)+np.arange(1, n_events+1)/10
    dt *= MaxDelay

    return dt

def times_from_fixed_rate(rate, n_events, n_simulated_events, offset=0):
    """
    Function to generate event times with a fixed rate.

    The event times are drawn from a uniform distribution.
    For higher rates pile-up is possible. The normalization
    is achieved by a variable overall simulation length in time.

    !Rate normalization is only valid for one simualtion job!

    Args:
        rate (int or float): Mean event rate in Hz. 
        n_events (int): Number of events in file
        n_events_simulated (int): True number of events which were
            simulated.
        offset (int or float): Time offset to shift all event times to larger values.

    Returns:
        event_times (numpy.array): Array containing the start times of the events.
    """
    
    simulation_time = n_simulated_events/rate
    simulation_time *= 1e9
    
    event_times = np.sort(np.random.uniform(low=offset, high=simulation_time+offset, size=n_events))
    
    return event_times



import numpy as np
import warnings
import nestpy

print(f'Using nestpy version {nestpy.__version__}')

@np.vectorize
def quanta_from_NEST(en, model, e_field, A, Z, create_s2, **kwargs):
    """
    Function which uses NEST to yield photons and electrons
    for a given set of parameters.

    Note:
        In case the energy deposit is outside of the range of NEST a -1
        is returned.

    Args:
        en (numpy.array): Energy deposit of the interaction [keV]
        model (numpy.array): Nest Id for qunata generation (integers)
        e_field (numpy.array): Field value in the interaction site [V/cm]
        A (numpy.array): Atomic mass number
        Z (numpy.array): Atomic number
        create_s2 (bool): Specifies if S2 can be produced by interaction,
            in this case electrons are generated.
        kwargs: Additional keyword arguments which can be taken by
            GetYields e.g. density.

    Returns:
        photons (numpy.array): Number of generated photons
        electrons (numpy.array): Number of generated electrons
    """
    nc = nestpy.NESTcalc(nestpy.VDetector())
    density = 2.862  # g/cm^3

    # Fix for Kr83m events.
    # Energies have to be exactly 32.1 keV or 9.1 keV
    # See: https://github.com/NESTCollaboration/nest/blob/master/src/NEST.cpp#L567
    # and: https://github.com/NESTCollaboration/nest/blob/master/src/NEST.cpp#L585
    # TODO: There must be something odd with our clustering, that we need
    #  this?
    max_allowed_energy_difference = 1  # keV
    if model == 11:
        if abs(en - 32.1) > max_allowed_energy_difference:
            en = 32.1
        if abs(en - 9.1) > max_allowed_energy_difference:
            en = 9.1

    # Some addition taken from
    # https://github.com/NESTCollaboration/nestpy/blob/e82c71f864d7362fee87989ed642cd875845ae3e/src/nestpy/helpers.py#L94-L100
    if model == 0 and en > 2e2:
        warnings.warn(f"Energy deposition of {en} keV beyond NEST validity for NR model of 200 keV - Remove Interaction")
        return -1, -1
    if model == 7 and en > 3e3:
        warnings.warn(f"Energy deposition of {en} keV beyond NEST validity for gamma model of 3 MeV - Remove Interaction")
        return -1, -1
    if model == 8 and en > 3e3:
        warnings.warn(f"Energy deposition of {en} keV beyond NEST validity for beta model of 3 MeV - Remove Interaction")
        return -1, -1

    y = nc.GetYields(interaction=nestpy.INTERACTION_TYPE(model),
                     energy=en,
                     drift_field=e_field,
                     A=A,
                     Z=Z,
                     **kwargs
                     )

    event_quanta = nc.GetQuanta(y)  # Density argument is not use in function...

    photons = event_quanta.photons

    electrons = 0
    if create_s2:
        electrons = event_quanta.electrons

    return photons, electrons



import numba
import epix

# Fixed detector dimensions of XENONnT:
# See also: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:coordinate_system
xenonnt_sensitive_volume_radius = 66.4  # cm
xenonnt_z_gate_mesh = 0.  # bottom of the gate electrode
xenonnt_z_top_pmts = 7.3936  # cm
xenonnt_z_cathode = -148.6515  # cm ... top of the cathode electrode
xenonnt_z_bottom_pmts = -154.6555  # cm ... top surface of the bottom PMT window
xenonnt_z_lxe = 0.416 # cm ... liquid-gas interface

# Fixed detector dimensions of XENON1T:
# See also: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenon1t:analysis:coordinate_system
xenon1t_sensitive_volume_radius = 47.925  # cm
xenon1t_z_gate_mesh = 0.  # bottom of the gate electrode
xenon1t_z_top_pmts = 7.38  # cm
xenon1t_z_cathode = -96.87  # cm ... top of the cathode electrode
xenon1t_z_bottom_pmts = -103.09  # cm ... top surface of the bottom PMT window
xenon1t_z_lxe = 0.27 # cm ... liquid-gas interface


def xenonnt():
    """
    Default XENONnT TPC with different sensitive volumes.

    The structure for each volume is as follows:
        key: VolumeName
        items: keyword arguments for the Sensitive Volume class

    :return: A list of volumes for the default nT detector.
    """
    # TODO: Can we do this differently? Ideas are welcome....
    # Outer edges of the volume of interest. This is needed for a
    # preselection of interactions inside the desired volumes before
    # clustering.
    outer_cylinder = {'max_z': xenonnt_z_top_pmts,
                      'min_z': xenonnt_z_bottom_pmts,
                      'max_r': xenonnt_sensitive_volume_radius
                      }

    volumes = {'TPC': {'volume_id': 1,
                       'roi': _make_roi_cylinder(xenonnt_z_cathode,
                                                 xenonnt_z_gate_mesh,
                                                 xenonnt_sensitive_volume_radius),
                       'electric_field': 200,
                       'create_S2': True,
                       'xe_density': 2.862,
                       'efield_outside_map': 200,
                       'to_be_stored': True,
                       },
               'BelowCathode': {'volume_id': 2,
                                'roi': _make_roi_cylinder(xenonnt_z_bottom_pmts,
                                                          xenonnt_z_cathode,
                                                          xenonnt_sensitive_volume_radius),
                                'electric_field': 200,
                                'create_S2': False,
                                'xe_density': 2.862,
                                'efield_outside_map': 200,
                                'to_be_stored': True,
                                },
               'GasPhase': {'volume_id': 3,
                            'roi': _make_roi_cylinder(xenonnt_z_lxe,
                                                      xenonnt_z_top_pmts,
                                                      xenonnt_sensitive_volume_radius),
                            'electric_field': 200,
                            'create_S2': False,
                            'xe_density': 0.0177,
                            'efield_outside_map': 200,
                            'to_be_stored': False,
                            },
               }
    return volumes, outer_cylinder


def xenon1t():
    """
    Default XENON1T TPC with different sensitive volumes.

    The structure for each volume is as follows:
        key: VolumeName
        items: keyword arguments for the Sensitive Volume class

    :return: A list of volumes for the default 1T detector.
    """
    # TODO: Can we do this differently? Ideas are welcome....
    # Outer edges of the volume of interest. This is needed for a
    # preselection of interactions inside the desired volumes before
    # clustering.
    outer_cylinder = {'max_z': xenon1t_z_top_pmts,
                      'min_z': xenon1t_z_bottom_pmts,
                      'max_r': xenon1t_sensitive_volume_radius
                      }

    volumes = {'TPC': {'volume_id': 1,
                       'roi': _make_roi_cylinder(xenon1t_z_cathode,
                                                 xenon1t_z_gate_mesh,
                                                 xenon1t_sensitive_volume_radius),
                       'electric_field': 82,
                       'create_S2': True,
                       'xe_density': 2.862,
                       'efield_outside_map': 82,
                       'to_be_stored': True,
                       },
               'BelowCathode': {'volume_id': 2,
                                'roi': _make_roi_cylinder(xenon1t_z_bottom_pmts,
                                                          xenon1t_z_cathode,
                                                          xenon1t_sensitive_volume_radius),
                                'electric_field': 82,
                                'create_S2': False,
                                'xe_density': 2.862,
                                'efield_outside_map': 82,
                                'to_be_stored': True,
                                },
               'GasPhase': {'volume_id': 3,
                            'roi': _make_roi_cylinder(xenon1t_z_lxe,
                                                      xenon1t_z_top_pmts,
                                                      xenon1t_sensitive_volume_radius),
                            'electric_field': 82,
                            'create_S2': False,
                            'xe_density': 0.0177,
                            'efield_outside_map': 82,
                            'to_be_stored': False,
                            },
               }
    return volumes, outer_cylinder


def _make_roi_cylinder(z_min, z_max, r_max):
    """
    Function to generate different rois, this is needed since
    we would like to have numba.njitted functions later.
    """

    # see: https://numba.pydata.org/numba-doc/dev/user/faq.html
    @numba.njit()
    def roi(x, y, z):
        return epix.in_cylinder(x, y, z, z_min, z_max, r_max)

    return roi
